#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#pragma hdrstop
#include "bfnt.h"

#define TRUE 1
#define FALSE 0

BFNT* BfntOpenfp( FILE * fp, const char * mode ) {
	if ( strcmp(mode, "r") && strcmp(mode,"rb") )
		return NULL;
	if ( !fp )
		return NULL;
	BfntHeader header;
	long pos = ftell(fp);
	if ( fread( &header, sizeof header, 1, fp ) < 1 ) {
		fseek(fp, pos, SEEK_SET);
		return 0;
	}
	if ( memcmp(header.id, "BFNT\x1a", 5) ) {
		fseek(fp, pos, SEEK_SET);
		return 0;
	}
	BFNT * b = new BFNT;
	if ( !b ) {
		fseek(fp, pos, SEEK_SET);
		return 0;
	}
	b->tcolor = 0;
	if ( header.extSize ) {
		unsigned char * buf = (unsigned char *)malloc(header.extSize);
		fread(buf, header.extSize, 1, fp );
		int hsiz = header.hdrSize ;
		int i = -2;
		while ( i < header.extSize ) {
			switch ( buf[i+2] ) {
			case 0x10:	/* pbgw */
				b->tcolor = (buf[i + 3] << 16) + (buf[i + 4] << 8) + buf[i + 5];
				break;
			case 0x00: /* gtHg */
			case 0x03: /* 쐬 */
			case 0x3f: /* Rg */
				break;
			}
			i += hsiz;
			hsiz = (buf[i] + (buf[i+1]<<8));
		}
		free(buf);
	}
	b->num_pal = 0;
	b->pal = 0;
	b->num_pal = 1 << ((header.col & 0x7f) + 1);
	if ( header.col & 0x80 ) {
		/* palette */
		int npal = b->num_pal;
		int i;
		b->pal = (RGBQUAD *)malloc( npal * sizeof (RGBQUAD) );
		for ( i = 0 ; i < npal ; ++i ) {
			unsigned char buf[3];
			fread( buf, 3, 1, fp );
			b->pal[i].rgbBlue = buf[0];
			b->pal[i].rgbRed = buf[1];
			b->pal[i].rgbGreen = buf[2];
			b->pal[i].rgbReserved = 0;
		}
	} else if ( header.col == 0 ) {
		b->pal = (RGBQUAD *)malloc( 2 * sizeof (RGBQUAD) );
		b->pal[0].rgbBlue = 0;
		b->pal[0].rgbRed = 0;
		b->pal[0].rgbGreen = 0;
		b->pal[0].rgbReserved = 0;
		b->pal[1].rgbBlue = 255;
		b->pal[1].rgbRed = 255;
		b->pal[1].rgbGreen = 255;
		b->pal[1].rgbReserved = 0;
	}

	b->header = header;
	b->fp = fp;
	b->current_char = header.START;
	b->current_line = 0;
	return b;
}

void BfntClose( BFNT* b ) {
	if ( b ) {
		if ( b->pal ) free(b->pal);
		fclose(b->fp);
		delete b;
	}
}

int BfntEOL( BFNT* b ) {
	if ( b->current_line >= b->header.Ydots )
		return TRUE;
	return FALSE;
}
int BfntEOF( BFNT* b ) {
	if ( b->current_char > b->header.END )
		return TRUE;
	return FALSE;
}

int BfntReadLineByte( BFNT* b, unsigned char * buf, int startx, int pixels ) {
	int x;
	int pdiv = 1;
	int pmask = b->num_pal - 1;
	int pshift = 0;
	if ( BfntEOF(b) ) {
		printf( "? bfnteof\n" );
		return -1;
	}
	switch ( b->num_pal ) {
	case 2:
		pdiv = 8;
		pshift = 1;
		break;
	case 4:
		pdiv = 4;
		pshift = 2;
		break;
	case 8:
		pdiv = 2;
		pshift = 4;
		break;
	case 16:
		pdiv = 2;
		pshift = 4;
		break;
	case 32:
	case 64:
	case 128:
	case 256:
		pdiv = 1;
		pshift = 0;
		break;
	default:
		printf( "? color %d\n", b->num_pal );
		return 0;
	}
	if ( startx < 0 ) {
		pixels += startx;
		startx = 0;
	}
	if ( startx >= b->header.Xdots || pixels <= 0 ) {
		printf( "? startx\n" );
		return 0;
	}
	if ( startx+pixels > b->header.Xdots ) {
		pixels = b->header.Xdots - startx ;
	}
	int linebytes = (b->header.Xdots + pdiv - 1) / pdiv;
	char * loadbuf = (char *)malloc( linebytes );
	if ( fread( loadbuf, linebytes, 1, b->fp ) < 1 ) {
		printf( "? fread(%d:%d)\n",b->current_char,b->current_line );
		return -1;
	}
	for ( x = startx ; x < startx+pixels; ++x ) {
		buf[x - startx] = (unsigned char)((loadbuf[x / pdiv] >> ((pdiv - (x % pdiv) - 1) * pshift)) & pmask);
	}
	if ( ++(b->current_line) >= b->header.Ydots ) {
		b->current_line = 0;
		++(b->current_char);
	}
	return pixels;
}
